/*  Screem:  support.c,
 *  some useful functions
 *
 *  Copyright (C) 1999, 2000  David A Knight
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  For contact information with the author of this source code please see
 *  the AUTHORS file.  If there is no AUTHORS file present then check the
 *  about box under the help menu for a contact address
 */

#include <config.h>
#include <errno.h>

#include <unistd.h>

#include <gdk/gdkkeysyms.h>

#include <gtk/gtkliststore.h>

#include <stdio.h>
#include <string.h>

#include "support.h"
#include "readtags.h"

static gchar *screem_support_escape_regexp( const gchar *pattern );

/**
 * change_state:
 *
 * toggles widget sensitivity
 *
 * return values: none
 */
void change_state( GtkWidget *widget, gpointer data )
{
        gtk_widget_set_sensitive( widget, !GTK_WIDGET_IS_SENSITIVE( widget ) );
}

void change_visible( GtkWidget *widget, gpointer data )
{
	if( GTK_WIDGET_VISIBLE( widget ) ) {
		gtk_widget_hide( widget );
	} else {
		gtk_widget_show( widget );
	}
}

gchar * convert_keysym_state_to_string( guint keysym, guint state )
{
        gchar *key;
	GString *str;

        if( keysym == GDK_VoidSymbol || keysym == GDK_BackSpace ||
	    keysym == GDK_Delete || keysym == 0 ) {
                return NULL;
	}

        gdk_error_trap_push();
        key = gdk_keyval_name( keysym );
        gdk_error_trap_pop();
        if( ! key )
		return NULL;

	str = g_string_new( "" );

        if( state & GDK_CONTROL_MASK ) {
		g_string_append( str, "Control-" );
	}
        if( state & GDK_LOCK_MASK ) {
		g_string_append( str, "Lock-" );
        }
        if( state & GDK_SHIFT_MASK ) {
		g_string_append( str, "Shift-" );
        }
        if( state & GDK_MOD1_MASK ) {
		g_string_append( str, "Mod1-" );
        }
        if( state & GDK_MOD2_MASK ) {
		g_string_append( str, "Mod2-" );
        }
        if( state & GDK_MOD3_MASK ) {
		g_string_append( str, "Mod3-" );
        }
        if( state & GDK_MOD4_MASK ) {
		g_string_append( str, "Mod4-" );
        }
        if( state & GDK_MOD5_MASK ) {
		g_string_append( str, "Mod5-" );
        }

	if( str->len ) {
		g_string_append( str, key );
	}

	key = str->str;
	g_string_free( str, FALSE );

	return key;
}

gchar *screem_support_ctags( const gchar *dirname,
			     const gchar *word,
			     gchar **pattern )
{
	gchar *ret;
	tagFile *file;
	tagFileInfo *info;
	tagEntry *entry;

	gchar *tagfile;

	if( ! word ) {
		return NULL;
	}
	
	info = g_new0( tagFileInfo, 1 );
	entry = g_new0( tagEntry, 1 );

	ret = NULL;

	tagfile = g_strconcat( dirname, G_DIR_SEPARATOR_S, "tags", NULL );
	file = tagsOpen( tagfile, info );
	if( ! file ) {
		g_free( tagfile );
		tagfile = g_strconcat( dirname, G_DIR_SEPARATOR_S, "TAGS",
				       NULL );
		file = tagsOpen( tagfile, info );
	}
	g_free( tagfile );

	if( file ) {
		if(tagsFind(file, entry, word, 
			    TAG_OBSERVECASE | TAG_FULLMATCH) == TagSuccess) {
			ret = g_strconcat( dirname, G_DIR_SEPARATOR_S,
					   entry->file, NULL );
			if( pattern ) {
				if( entry->address.pattern ) {
					*pattern = screem_support_escape_regexp( entry->address.pattern );				
				} else {
					*pattern = NULL;
				}
			}
		}
	}
	
	g_free( entry );
	g_free( info );

	return ret;
}

/* process regexp, strip leading and trailing '/' characters,
   escape ( ) { } */
static gchar *screem_support_escape_regexp( const gchar *pattern )
{
	gchar *ret;
	GString *temp;

	g_return_val_if_fail( pattern != NULL, NULL );
	g_return_val_if_fail( *pattern == '/', NULL );

	temp = g_string_new( "" );
	
	pattern ++;
	while( *pattern != '\0' &&
	       ( *pattern != '/' || *( pattern - 1 ) == '\\' ) ) {
		gchar p;
		gchar c;

		p = *( pattern - 1 );
		c = *pattern;

		if( p != '\\' &&
		    ( c == '(' || c == ')' || c == '{' || c == '}' ) ) {
			g_string_append_c( temp, '\\' );
			g_string_append_c( temp, c );
		} else {
			g_string_append_c( temp, c );
		}

		pattern ++;
	}

	if( temp->len ) {
		ret = g_strdup( temp->str );
	} else {
		ret = NULL;
	}
	g_string_free( temp, TRUE );
	

	return ret;
}


GtkTreeIter *screem_support_find_in_list( GtkListStore *store, guint column, const gchar *str )
{
	GtkTreeIter it;
	GtkTreeIter *ret;
	gchar *val;
	
	ret = NULL;

	g_assert( gtk_tree_model_get_column_type( GTK_TREE_MODEL( store ),
						  column ) == G_TYPE_STRING );
	
	if( gtk_tree_model_get_iter_first( GTK_TREE_MODEL( store ), &it ) ) {
		do {
			gtk_tree_model_get( GTK_TREE_MODEL( store ), &it, 
					    column, &val, -1 );

			if( val && ! strncmp( "regex:", val, strlen( "regex:" ) ) )  {
				/* regexp pattern */
				g_warning( "regex: pattern match not supported yet\n" );
			} else if( val && ! strncmp( "glob:", val, strlen( "glob:" ) ) ) {
				/* glob */
				if( fnmatch( val + strlen( "glob:" ), str, 0 ) == 0 ) {
					ret = gtk_tree_iter_copy( &it );
				}
			} else 	if( val && ! strcmp( val, str ) ) {
				/* string pattern */
				
				ret = gtk_tree_iter_copy( &it );
			}
			g_free( val );
		} while( ( ! ret ) && 
			 gtk_tree_model_iter_next( GTK_TREE_MODEL( store ),
				 		   &it ) );
	}
	
	return ret;
}

gchar *screem_support_charset_convert( const gchar *data )
{
	const gchar *end;
	gchar *ret;
	guint len;
	
	ret = NULL;
	if( data && ! g_utf8_validate( data, strlen( data ), &end ) ) {
		gchar *temp = NULL;
		const gchar *charset;
		gboolean utf8;
		const char *encodings_to_try[ 2 ] = { 0, 0 };

		utf8 = g_get_charset( &charset );

		if( ! utf8 ) {
			encodings_to_try[ 0 ] = charset;
		}
		if( g_ascii_strcasecmp( charset, "ISO-8859-1" ) ) {
			encodings_to_try[ 1 ] = "ISO-8859-1";
		}

		if( encodings_to_try[ 0 ] ) {
			temp = g_convert( data, strlen( data ), "UTF-8", 
					  encodings_to_try[ 0 ], 
					  NULL, &len, NULL );
		}
		if( ! temp && encodings_to_try[ 1 ] ) {
			temp = g_convert( data, strlen( data ), "UTF-8", 
					  encodings_to_try[ 1 ], 
					  NULL, &len, NULL );
		}

		if( temp ) {
			data = temp;
		}
		ret = data;
	} else if( data ) {
		ret = g_strdup( data );
	}

	return ret;
}
